#include "gui_fsm.hpp"
#include <iostream>
#include <chrono>
#include <numeric>
#include <functional>
#include "components.hpp"
#include <numbers>
#include "systems.hpp"
#include "events.hpp"

namespace gui {
  using namespace components;

  FSM::FSM() :
    $window(sf::VideoMode({SCREEN_WIDTH, SCREEN_HEIGHT}), "Zed's Raycaster Thing"),
    $main_ui($window, $levels.current(), $textures),
    $renderer($window),
    $level($levels.current()),
    $map_ui($level),
    $combat_ui($level),
    $status_ui($level),
    $font{FONT_FILE_NAME},
    $rayview($textures, RAY_VIEW_WIDTH, RAY_VIEW_HEIGHT)
    {
      $textures.load_tiles();
      $textures.load_sprites();
    }

  void FSM::event(Event ev) {
    switch($state) {
      FSM_STATE(State, START, ev);
      FSM_STATE(State, MOVING, ev);
      FSM_STATE(State, ATTACKING, ev);
      FSM_STATE(State, MAPPING, ev);
      FSM_STATE(State, ROTATING, ev);
      FSM_STATE(State, IDLE, ev);
      FSM_STATE(State, IN_COMBAT, ev);
      FSM_STATE(State, COMBAT_ROTATE, ev);
      FSM_STATE(State, END, ev);
    }
  }

  void FSM::START(Event ) {
    generate_map();
    $level.world->set_the<Debug>({});
    $rayview.init_shaders();
    $rayview.set_position(RAY_VIEW_X, RAY_VIEW_Y);
    $rayview.position_camera($player.x + 0.5, $player.y + 0.5);

    $main_ui.render();
    $combat_ui.render($textures);
    $status_ui.render($textures);
    $status_ui.log("Welcome to the game!");

    $renderer.init_terminal();
    $map_ui.create_render();
    $map_ui.resize_canvas();
    $renderer.resize_grid(MAX_FONT_SIZE, $map_ui);

    run_systems();
    state(State::IDLE);
  }

  void FSM::MAPPING(Event ev) {
    using enum Event;
    // BUG: can't close window when in mapping
    switch(ev) {
      case MAP_OPEN:
        state(State::IDLE);
        break;
      case CLOSE:
        state(State::IDLE);
        break;
      case TICK:
        break;
      default:
        dbc::log("invalid event sent to MAPPING");
    }
  }

  void FSM::MOVING(Event ) {
    if($camera.play_move($rayview)) {
      System::plan_motion(*$level.world, {size_t($camera.target_x), size_t($camera.target_y)});
      run_systems();
      state(State::IDLE);
    }
  }

  void FSM::ATTACKING(Event ev) {
    using enum Event;
    switch(ev) {
      case TICK: {
        System::combat($level);
        run_systems();
        state(State::IN_COMBAT);
       } break;
      case STOP_COMBAT:
        dbc::log("Exiting ATTACKING STATE");
        state(State::IDLE);
        break;
      case ATTACK:
        // ignore these since they're just from SFML not having discrete events
        break;
      default:
        dbc::log(fmt::format("In ATTACKING state, unhandled event {}", (int)ev));
    }
  }

  void FSM::ROTATING(Event ) {
    if($camera.play_rotate($rayview)) {
      state(State::IDLE);
    }
  }

  void FSM::COMBAT_ROTATE(Event ) {
    if($camera.play_rotate($rayview)) {
      state(State::IN_COMBAT);
    }
  }

  void FSM::IDLE(Event ev) {
    using enum Event;

    switch(ev) {
      case QUIT:
        $window.close();
        state(State::END);
        return; // done
      case MOVE_FORWARD:
        try_move(1, false);
        break;
      case MOVE_BACK:
        try_move(-1, false);
        break;
      case MOVE_LEFT:
        try_move(1, true);
        break;
      case MOVE_RIGHT:
        try_move(-1, true);
        break;
      case ROTATE_LEFT:
        $camera.plan_rotate($rayview, 1);
        state(State::ROTATING);
        break;
      case ROTATE_RIGHT:
        $camera.plan_rotate($rayview, -1);
        state(State::ROTATING);
        break;
      case MAP_OPEN:
        $renderer.resize_grid(MAX_FONT_SIZE, $map_ui);
        $map_ui.resize_canvas();
        state(State::MAPPING);
        break;
      case ATTACK:
        state(State::ATTACKING);
        break;
      case START_COMBAT:
        $main_ui.$overlay_ui.show_sprite("top_right", "cinqueda");
        state(State::IN_COMBAT);
        break;
      case CLOSE:
        dbc::log("Nothing to close.");
        break;
      case STOP_COMBAT:
      case TICK:
        // do nothing
        break;
      default:
        dbc::sentinel("unhandled event in IDLE");
    }
  }

  void FSM::IN_COMBAT(Event ev) {
    using enum Event;

    switch(ev) {
      case ATTACK:
        $status_ui.log("You attack!");
        state(State::ATTACKING);
        break;
      case ROTATE_LEFT:
        $camera.plan_rotate($rayview, 1);
        state(State::COMBAT_ROTATE);
        break;
      case ROTATE_RIGHT:
        $camera.plan_rotate($rayview, -1);
        state(State::COMBAT_ROTATE);
        break;
      case STOP_COMBAT:
        $main_ui.$overlay_ui.close_sprite("top_right");
        state(State::IDLE);
        break;
      case QUIT:
        $window.close();
        state(State::END);
        return;
      default:
        break;
    }
  }

  void FSM::try_move(int dir, bool strafe) {
    using enum State;
    // prevent moving into occupied space
    Point move_to = $camera.plan_move($rayview, dir, strafe);

    if($level.map->can_move(move_to) && !$level.collision->occupied(move_to)) {
      state(MOVING);
    } else {
      state(IDLE);
      $camera.abort_plan($rayview);
    }
  }

  void FSM::END(Event ev) {
    fmt::println("END: received event after done: {}", int(ev));
  }

  void FSM::keyboard() {
    while(const auto keyev = $window.pollEvent()) {
      if(keyev->is<sf::Event::Closed>()) {
        event(Event::QUIT);
      }

      if(const auto* key = keyev->getIf<sf::Event::KeyPressed>()) {
        using KEY = sf::Keyboard::Scan;
        switch(key->scancode) {
          case KEY::W:
            event(Event::MOVE_FORWARD);
            break;
          case KEY::S:
            event(Event::MOVE_BACK);
            break;
          case KEY::Q:
            event(Event::ROTATE_LEFT);
            break;
          case KEY::E:
            event(Event::ROTATE_RIGHT);
            break;
          case KEY::D:
            event(Event::MOVE_RIGHT);
            break;
          case KEY::A:
            event(Event::MOVE_LEFT);
            break;
          case KEY::R:
            $main_ui.$stats.reset();
            break;
          case KEY::M:
            event(Event::MAP_OPEN);
            break;
          case KEY::Escape:
            event(Event::CLOSE);
            break;
          case KEY::P:
            $main_ui.debug();
            break;
          default:
            break; // ignored
        }
      }
    }
  }


  void FSM::draw_gui() {
    $status_ui.draw($window);
    $combat_ui.draw($window);
    $main_ui.draw();
  }

  void FSM::render() {
    if(in_state(State::MAPPING)) {
      $window.clear();
      $map_ui.render();
      $renderer.draw($map_ui);
    } else {
      auto start = std::chrono::high_resolution_clock::now();
      $rayview.draw($window);
      auto end = std::chrono::high_resolution_clock::now();
      auto elapsed = std::chrono::duration<double>(end - start);
      $main_ui.$stats.sample(1/elapsed.count());

      draw_gui();
      $main_ui.draw_blood();
    }

    $window.display();
  }

  void FSM::mouse() {
    if(sf::Mouse::isButtonPressed(sf::Mouse::Button::Left)) {
      sf::Vector2f pos = $window.mapPixelToCoords(sf::Mouse::getPosition($window));
      $combat_ui.$gui.mouse(pos.x, pos.y);
      $status_ui.$gui.mouse(pos.x, pos.y);
    }
  }

  void FSM::generate_map() {
    // ZED: this should eventually go away now that level manager is in play
    auto& player = $level.world->get_the<Player>();
    auto& player_position = $level.world->get<Position>(player.entity);
    $player = player_position.location;
    $rayview.set_level($level);
  }

  void FSM::run_systems() {
    System::enemy_pathing($level);
    System::collision($level);
    System::motion($level);
    System::lighting($level);
    System::death($level, $levels.$components);
  }

  bool FSM::active() {
    return !in_state(State::END);
  }

  void FSM::handle_world_events() {
    using eGUI = Events::GUI;
    auto& world = *$level.world;

    while(world.has_event<eGUI>()) {
      auto [evt, entity, data] = world.recv<eGUI>();
      auto player = world.get_the<Player>();

      switch(evt) {
        case eGUI::COMBAT: {
            auto &damage = std::any_cast<Events::Combat&>(data);

            if(damage.enemy_did > 0) {
              $status_ui.log(fmt::format("Enemy HIT YOU for {} damage!", damage.enemy_did));
              auto player_combat = world.get<Combat>(player.entity);
              $combat_ui.set_damage(float(player_combat.hp) / float(player_combat.max_hp));
            } else {
              $status_ui.log("Enemy MISSED YOU.");
            }

            if(damage.player_did > 0) {
              $status_ui.log(fmt::format("You HIT enemy for {} damage!", damage.player_did));
            } else {
              $status_ui.log("You MISSED the enemy.");
            }
          }
          break;
        case eGUI::COMBAT_START:
          event(Event::START_COMBAT);
          break;
        case eGUI::NO_NEIGHBORS:
          event(Event::STOP_COMBAT);
          break;
        case eGUI::LOOT: {
            // auto &item = std::any_cast<InventoryItem&>(data);
            // $status_ui.log(fmt::format("You picked up a {}.",
            //      std::string(item.data["name"])));
            $status_ui.log("You picked up an item.");
          }
          break;
        case eGUI::ATTACK:
          event(Event::ATTACK);
          break;
        case eGUI::DEATH: {
            if(entity != player.entity) {
              auto &sprite = $level.world->get<Sprite>(entity);
              $rayview.update_sprite(entity, sprite);
            }
          } break;
        case eGUI::NOOP:
          $status_ui.log(fmt::format("NOOP EVENT! {},{}", evt, entity));
          break;
        default:
          $status_ui.log(fmt::format("INVALID EVENT! {},{}", evt, entity));
      }
    }
  }
}
